try:
    import unittest2 as unittest
except ImportError:
    import unittest

from pamqp import specification


class ArgumentTypeTests(unittest.TestCase):
    def test_basic_ack_has_delivery_tag(self):
        self.assertEqual(specification.Basic.Ack.type('delivery_tag'),
                         'longlong')

    def test_basic_ack_has_multiple(self):
        self.assertEqual(specification.Basic.Ack.type('multiple'), 'bit')

    def test_basic_cancel_has_consumer_tag(self):
        self.assertEqual(specification.Basic.Cancel.type('consumer_tag'),
                         'shortstr')

    def test_basic_cancel_has_nowait(self):
        self.assertEqual(specification.Basic.Cancel.type('nowait'), 'bit')

    def test_basic_cancelok_has_consumer_tag(self):
        self.assertEqual(specification.Basic.CancelOk.type('consumer_tag'),
                         'shortstr')

    def test_basic_consume_has_ticket(self):
        self.assertEqual(specification.Basic.Consume.type('ticket'), 'short')

    def test_basic_consume_has_queue(self):
        self.assertEqual(specification.Basic.Consume.type('queue'), 'shortstr')

    def test_basic_consume_has_consumer_tag(self):
        self.assertEqual(specification.Basic.Consume.type('consumer_tag'),
                         'shortstr')

    def test_basic_consume_has_no_local(self):
        self.assertEqual(specification.Basic.Consume.type('no_local'), 'bit')

    def test_basic_consume_has_no_ack(self):
        self.assertEqual(specification.Basic.Consume.type('no_ack'), 'bit')

    def test_basic_consume_has_exclusive(self):
        self.assertEqual(specification.Basic.Consume.type('exclusive'), 'bit')

    def test_basic_consume_has_nowait(self):
        self.assertEqual(specification.Basic.Consume.type('nowait'), 'bit')

    def test_basic_consume_has_arguments(self):
        self.assertEqual(specification.Basic.Consume.type('arguments'),
                         'table')

    def test_basic_consumeok_has_consumer_tag(self):
        self.assertEqual(specification.Basic.ConsumeOk.type('consumer_tag'),
                         'shortstr')

    def test_basic_deliver_has_consumer_tag(self):
        self.assertEqual(specification.Basic.Deliver.type('consumer_tag'),
                         'shortstr')

    def test_basic_deliver_has_delivery_tag(self):
        self.assertEqual(specification.Basic.Deliver.type('delivery_tag'),
                         'longlong')

    def test_basic_deliver_has_redelivered(self):
        self.assertEqual(specification.Basic.Deliver.type('redelivered'),
                         'bit')

    def test_basic_deliver_has_exchange(self):
        self.assertEqual(specification.Basic.Deliver.type('exchange'),
                         'shortstr')

    def test_basic_deliver_has_routing_key(self):
        self.assertEqual(specification.Basic.Deliver.type('routing_key'),
                         'shortstr')

    def test_basic_get_has_ticket(self):
        self.assertEqual(specification.Basic.Get.type('ticket'), 'short')

    def test_basic_get_has_queue(self):
        self.assertEqual(specification.Basic.Get.type('queue'), 'shortstr')

    def test_basic_get_has_no_ack(self):
        self.assertEqual(specification.Basic.Get.type('no_ack'), 'bit')

    def test_basic_getempty_has_cluster_id(self):
        self.assertEqual(specification.Basic.GetEmpty.type('cluster_id'),
                         'shortstr')

    def test_basic_getok_has_delivery_tag(self):
        self.assertEqual(specification.Basic.GetOk.type('delivery_tag'),
                         'longlong')

    def test_basic_getok_has_redelivered(self):
        self.assertEqual(specification.Basic.GetOk.type('redelivered'), 'bit')

    def test_basic_getok_has_exchange(self):
        self.assertEqual(specification.Basic.GetOk.type('exchange'),
                         'shortstr')

    def test_basic_getok_has_routing_key(self):
        self.assertEqual(specification.Basic.GetOk.type('routing_key'),
                         'shortstr')

    def test_basic_getok_has_message_count(self):
        self.assertEqual(specification.Basic.GetOk.type('message_count'),
                         'long')

    def test_basic_nack_has_delivery_tag(self):
        self.assertEqual(specification.Basic.Nack.type('delivery_tag'),
                         'longlong')

    def test_basic_nack_has_multiple(self):
        self.assertEqual(specification.Basic.Nack.type('multiple'), 'bit')

    def test_basic_nack_has_requeue(self):
        self.assertEqual(specification.Basic.Nack.type('requeue'), 'bit')

    def test_basic_publish_has_ticket(self):
        self.assertEqual(specification.Basic.Publish.type('ticket'), 'short')

    def test_basic_publish_has_exchange(self):
        self.assertEqual(specification.Basic.Publish.type('exchange'),
                         'shortstr')

    def test_basic_publish_has_routing_key(self):
        self.assertEqual(specification.Basic.Publish.type('routing_key'),
                         'shortstr')

    def test_basic_publish_has_mandatory(self):
        self.assertEqual(specification.Basic.Publish.type('mandatory'), 'bit')

    def test_basic_publish_has_immediate(self):
        self.assertEqual(specification.Basic.Publish.type('immediate'), 'bit')

    def test_basic_qos_has_prefetch_size(self):
        self.assertEqual(specification.Basic.Qos.type('prefetch_size'), 'long')

    def test_basic_qos_has_prefetch_count(self):
        self.assertEqual(specification.Basic.Qos.type('prefetch_count'),
                         'short')

    def test_basic_qos_has_global_(self):
        self.assertEqual(specification.Basic.Qos.type('global_'), 'bit')

    def test_basic_recover_has_requeue(self):
        self.assertEqual(specification.Basic.Recover.type('requeue'), 'bit')

    def test_basic_recoverasync_has_requeue(self):
        self.assertEqual(specification.Basic.RecoverAsync.type('requeue'),
                         'bit')

    def test_basic_reject_has_delivery_tag(self):
        self.assertEqual(specification.Basic.Reject.type('delivery_tag'),
                         'longlong')

    def test_basic_reject_has_requeue(self):
        self.assertEqual(specification.Basic.Reject.type('requeue'), 'bit')

    def test_basic_return_has_reply_code(self):
        self.assertEqual(specification.Basic.Return.type('reply_code'),
                         'short')

    def test_basic_return_has_reply_text(self):
        self.assertEqual(specification.Basic.Return.type('reply_text'),
                         'shortstr')

    def test_basic_return_has_exchange(self):
        self.assertEqual(specification.Basic.Return.type('exchange'),
                         'shortstr')

    def test_basic_return_has_routing_key(self):
        self.assertEqual(specification.Basic.Return.type('routing_key'),
                         'shortstr')

    def test_channel_close_has_reply_code(self):
        self.assertEqual(specification.Channel.Close.type('reply_code'),
                         'short')

    def test_channel_close_has_reply_text(self):
        self.assertEqual(specification.Channel.Close.type('reply_text'),
                         'shortstr')

    def test_channel_close_has_class_id(self):
        self.assertEqual(specification.Channel.Close.type('class_id'), 'short')

    def test_channel_close_has_method_id(self):
        self.assertEqual(specification.Channel.Close.type('method_id'),
                         'short')

    def test_channel_flow_has_active(self):
        self.assertEqual(specification.Channel.Flow.type('active'), 'bit')

    def test_channel_flowok_has_active(self):
        self.assertEqual(specification.Channel.FlowOk.type('active'), 'bit')

    def test_channel_open_has_out_of_band(self):
        self.assertEqual(specification.Channel.Open.type('out_of_band'),
                         'shortstr')

    def test_channel_openok_has_channel_id(self):
        self.assertEqual(specification.Channel.OpenOk.type('channel_id'),
                         'longstr')

    def test_confirm_select_has_nowait(self):
        self.assertEqual(specification.Confirm.Select.type('nowait'), 'bit')

    def test_connection_blocked_has_reason(self):
        self.assertEqual(specification.Connection.Blocked.type('reason'),
                         'shortstr')

    def test_connection_close_has_reply_code(self):
        self.assertEqual(specification.Connection.Close.type('reply_code'),
                         'short')

    def test_connection_close_has_reply_text(self):
        self.assertEqual(specification.Connection.Close.type('reply_text'),
                         'shortstr')

    def test_connection_close_has_class_id(self):
        self.assertEqual(specification.Connection.Close.type('class_id'),
                         'short')

    def test_connection_close_has_method_id(self):
        self.assertEqual(specification.Connection.Close.type('method_id'),
                         'short')

    def test_connection_open_has_virtual_host(self):
        self.assertEqual(specification.Connection.Open.type('virtual_host'),
                         'shortstr')

    def test_connection_open_has_capabilities(self):
        self.assertEqual(specification.Connection.Open.type('capabilities'),
                         'shortstr')

    def test_connection_open_has_insist(self):
        self.assertEqual(specification.Connection.Open.type('insist'), 'bit')

    def test_connection_openok_has_known_hosts(self):
        self.assertEqual(specification.Connection.OpenOk.type('known_hosts'),
                         'shortstr')

    def test_connection_secure_has_challenge(self):
        self.assertEqual(specification.Connection.Secure.type('challenge'),
                         'longstr')

    def test_connection_secureok_has_response(self):
        self.assertEqual(specification.Connection.SecureOk.type('response'),
                         'longstr')

    def test_connection_start_has_version_major(self):
        self.assertEqual(specification.Connection.Start.type('version_major'),
                         'octet')

    def test_connection_start_has_version_minor(self):
        self.assertEqual(specification.Connection.Start.type('version_minor'),
                         'octet')

    def test_connection_start_has_server_properties(self):
        self.assertEqual(
            specification.Connection.Start.type('server_properties'), 'table')

    def test_connection_start_has_mechanisms(self):
        self.assertEqual(specification.Connection.Start.type('mechanisms'),
                         'longstr')

    def test_connection_start_has_locales(self):
        self.assertEqual(specification.Connection.Start.type('locales'),
                         'longstr')

    def test_connection_startok_has_client_properties(self):
        self.assertEqual(
            specification.Connection.StartOk.type('client_properties'),
            'table')

    def test_connection_startok_has_mechanism(self):
        self.assertEqual(specification.Connection.StartOk.type('mechanism'),
                         'shortstr')

    def test_connection_startok_has_response(self):
        self.assertEqual(specification.Connection.StartOk.type('response'),
                         'longstr')

    def test_connection_startok_has_locale(self):
        self.assertEqual(specification.Connection.StartOk.type('locale'),
                         'shortstr')

    def test_connection_tune_has_channel_max(self):
        self.assertEqual(specification.Connection.Tune.type('channel_max'),
                         'short')

    def test_connection_tune_has_frame_max(self):
        self.assertEqual(specification.Connection.Tune.type('frame_max'),
                         'long')

    def test_connection_tune_has_heartbeat(self):
        self.assertEqual(specification.Connection.Tune.type('heartbeat'),
                         'short')

    def test_connection_tuneok_has_channel_max(self):
        self.assertEqual(specification.Connection.TuneOk.type('channel_max'),
                         'short')

    def test_connection_tuneok_has_frame_max(self):
        self.assertEqual(specification.Connection.TuneOk.type('frame_max'),
                         'long')

    def test_connection_tuneok_has_heartbeat(self):
        self.assertEqual(specification.Connection.TuneOk.type('heartbeat'),
                         'short')

    def test_exchange_bind_has_ticket(self):
        self.assertEqual(specification.Exchange.Bind.type('ticket'), 'short')

    def test_exchange_bind_has_destination(self):
        self.assertEqual(specification.Exchange.Bind.type('destination'),
                         'shortstr')

    def test_exchange_bind_has_source(self):
        self.assertEqual(specification.Exchange.Bind.type('source'),
                         'shortstr')

    def test_exchange_bind_has_routing_key(self):
        self.assertEqual(specification.Exchange.Bind.type('routing_key'),
                         'shortstr')

    def test_exchange_bind_has_nowait(self):
        self.assertEqual(specification.Exchange.Bind.type('nowait'), 'bit')

    def test_exchange_bind_has_arguments(self):
        self.assertEqual(specification.Exchange.Bind.type('arguments'),
                         'table')

    def test_exchange_declare_has_ticket(self):
        self.assertEqual(specification.Exchange.Declare.type('ticket'),
                         'short')

    def test_exchange_declare_has_exchange(self):
        self.assertEqual(specification.Exchange.Declare.type('exchange'),
                         'shortstr')

    def test_exchange_declare_has_exchange_type(self):
        self.assertEqual(specification.Exchange.Declare.type('exchange_type'),
                         'shortstr')

    def test_exchange_declare_has_passive(self):
        self.assertEqual(specification.Exchange.Declare.type('passive'), 'bit')

    def test_exchange_declare_has_durable(self):
        self.assertEqual(specification.Exchange.Declare.type('durable'), 'bit')

    def test_exchange_declare_has_auto_delete(self):
        self.assertEqual(specification.Exchange.Declare.type('auto_delete'),
                         'bit')

    def test_exchange_declare_has_internal(self):
        self.assertEqual(specification.Exchange.Declare.type('internal'),
                         'bit')

    def test_exchange_declare_has_nowait(self):
        self.assertEqual(specification.Exchange.Declare.type('nowait'), 'bit')

    def test_exchange_declare_has_arguments(self):
        self.assertEqual(specification.Exchange.Declare.type('arguments'),
                         'table')

    def test_exchange_delete_has_ticket(self):
        self.assertEqual(specification.Exchange.Delete.type('ticket'), 'short')

    def test_exchange_delete_has_exchange(self):
        self.assertEqual(specification.Exchange.Delete.type('exchange'),
                         'shortstr')

    def test_exchange_delete_has_if_unused(self):
        self.assertEqual(specification.Exchange.Delete.type('if_unused'),
                         'bit')

    def test_exchange_delete_has_nowait(self):
        self.assertEqual(specification.Exchange.Delete.type('nowait'), 'bit')

    def test_exchange_unbind_has_ticket(self):
        self.assertEqual(specification.Exchange.Unbind.type('ticket'), 'short')

    def test_exchange_unbind_has_destination(self):
        self.assertEqual(specification.Exchange.Unbind.type('destination'),
                         'shortstr')

    def test_exchange_unbind_has_source(self):
        self.assertEqual(specification.Exchange.Unbind.type('source'),
                         'shortstr')

    def test_exchange_unbind_has_routing_key(self):
        self.assertEqual(specification.Exchange.Unbind.type('routing_key'),
                         'shortstr')

    def test_exchange_unbind_has_nowait(self):
        self.assertEqual(specification.Exchange.Unbind.type('nowait'), 'bit')

    def test_exchange_unbind_has_arguments(self):
        self.assertEqual(specification.Exchange.Unbind.type('arguments'),
                         'table')

    def test_queue_bind_has_ticket(self):
        self.assertEqual(specification.Queue.Bind.type('ticket'), 'short')

    def test_queue_bind_has_queue(self):
        self.assertEqual(specification.Queue.Bind.type('queue'), 'shortstr')

    def test_queue_bind_has_exchange(self):
        self.assertEqual(specification.Queue.Bind.type('exchange'), 'shortstr')

    def test_queue_bind_has_routing_key(self):
        self.assertEqual(specification.Queue.Bind.type('routing_key'),
                         'shortstr')

    def test_queue_bind_has_nowait(self):
        self.assertEqual(specification.Queue.Bind.type('nowait'), 'bit')

    def test_queue_bind_has_arguments(self):
        self.assertEqual(specification.Queue.Bind.type('arguments'), 'table')

    def test_queue_declare_has_ticket(self):
        self.assertEqual(specification.Queue.Declare.type('ticket'), 'short')

    def test_queue_declare_has_queue(self):
        self.assertEqual(specification.Queue.Declare.type('queue'), 'shortstr')

    def test_queue_declare_has_passive(self):
        self.assertEqual(specification.Queue.Declare.type('passive'), 'bit')

    def test_queue_declare_has_durable(self):
        self.assertEqual(specification.Queue.Declare.type('durable'), 'bit')

    def test_queue_declare_has_exclusive(self):
        self.assertEqual(specification.Queue.Declare.type('exclusive'), 'bit')

    def test_queue_declare_has_auto_delete(self):
        self.assertEqual(specification.Queue.Declare.type('auto_delete'),
                         'bit')

    def test_queue_declare_has_nowait(self):
        self.assertEqual(specification.Queue.Declare.type('nowait'), 'bit')

    def test_queue_declare_has_arguments(self):
        self.assertEqual(specification.Queue.Declare.type('arguments'),
                         'table')

    def test_queue_declareok_has_queue(self):
        self.assertEqual(specification.Queue.DeclareOk.type('queue'),
                         'shortstr')

    def test_queue_declareok_has_message_count(self):
        self.assertEqual(specification.Queue.DeclareOk.type('message_count'),
                         'long')

    def test_queue_declareok_has_consumer_count(self):
        self.assertEqual(specification.Queue.DeclareOk.type('consumer_count'),
                         'long')

    def test_queue_delete_has_ticket(self):
        self.assertEqual(specification.Queue.Delete.type('ticket'), 'short')

    def test_queue_delete_has_queue(self):
        self.assertEqual(specification.Queue.Delete.type('queue'), 'shortstr')

    def test_queue_delete_has_if_unused(self):
        self.assertEqual(specification.Queue.Delete.type('if_unused'), 'bit')

    def test_queue_delete_has_if_empty(self):
        self.assertEqual(specification.Queue.Delete.type('if_empty'), 'bit')

    def test_queue_delete_has_nowait(self):
        self.assertEqual(specification.Queue.Delete.type('nowait'), 'bit')

    def test_queue_deleteok_has_message_count(self):
        self.assertEqual(specification.Queue.DeleteOk.type('message_count'),
                         'long')

    def test_queue_purge_has_ticket(self):
        self.assertEqual(specification.Queue.Purge.type('ticket'), 'short')

    def test_queue_purge_has_queue(self):
        self.assertEqual(specification.Queue.Purge.type('queue'), 'shortstr')

    def test_queue_purge_has_nowait(self):
        self.assertEqual(specification.Queue.Purge.type('nowait'), 'bit')

    def test_queue_purgeok_has_message_count(self):
        self.assertEqual(specification.Queue.PurgeOk.type('message_count'),
                         'long')

    def test_queue_unbind_has_ticket(self):
        self.assertEqual(specification.Queue.Unbind.type('ticket'), 'short')

    def test_queue_unbind_has_queue(self):
        self.assertEqual(specification.Queue.Unbind.type('queue'), 'shortstr')

    def test_queue_unbind_has_exchange(self):
        self.assertEqual(specification.Queue.Unbind.type('exchange'),
                         'shortstr')

    def test_queue_unbind_has_routing_key(self):
        self.assertEqual(specification.Queue.Unbind.type('routing_key'),
                         'shortstr')

    def test_queue_unbind_has_arguments(self):
        self.assertEqual(specification.Queue.Unbind.type('arguments'), 'table')


class AttributeInMethodTests(unittest.TestCase):
    def test_basic_ack_has_delivery_tag(self):
        self.assertIn('delivery_tag', specification.Basic.Ack())

    def test_basic_ack_has_multiple(self):
        self.assertIn('multiple', specification.Basic.Ack())

    def test_basic_cancel_has_consumer_tag(self):
        self.assertIn('consumer_tag', specification.Basic.Cancel())

    def test_basic_cancel_has_nowait(self):
        self.assertIn('nowait', specification.Basic.Cancel())

    def test_basic_cancelok_has_consumer_tag(self):
        self.assertIn('consumer_tag', specification.Basic.CancelOk())

    def test_basic_consume_has_ticket(self):
        self.assertIn('ticket', specification.Basic.Consume())

    def test_basic_consume_has_queue(self):
        self.assertIn('queue', specification.Basic.Consume())

    def test_basic_consume_has_consumer_tag(self):
        self.assertIn('consumer_tag', specification.Basic.Consume())

    def test_basic_consume_has_no_local(self):
        self.assertIn('no_local', specification.Basic.Consume())

    def test_basic_consume_has_no_ack(self):
        self.assertIn('no_ack', specification.Basic.Consume())

    def test_basic_consume_has_exclusive(self):
        self.assertIn('exclusive', specification.Basic.Consume())

    def test_basic_consume_has_nowait(self):
        self.assertIn('nowait', specification.Basic.Consume())

    def test_basic_consume_has_arguments(self):
        self.assertIn('arguments', specification.Basic.Consume())

    def test_basic_consumeok_has_consumer_tag(self):
        self.assertIn('consumer_tag', specification.Basic.ConsumeOk())

    def test_basic_deliver_has_consumer_tag(self):
        self.assertIn('consumer_tag', specification.Basic.Deliver())

    def test_basic_deliver_has_delivery_tag(self):
        self.assertIn('delivery_tag', specification.Basic.Deliver())

    def test_basic_deliver_has_redelivered(self):
        self.assertIn('redelivered', specification.Basic.Deliver())

    def test_basic_deliver_has_exchange(self):
        self.assertIn('exchange', specification.Basic.Deliver())

    def test_basic_deliver_has_routing_key(self):
        self.assertIn('routing_key', specification.Basic.Deliver())

    def test_basic_get_has_ticket(self):
        self.assertIn('ticket', specification.Basic.Get())

    def test_basic_get_has_queue(self):
        self.assertIn('queue', specification.Basic.Get())

    def test_basic_get_has_no_ack(self):
        self.assertIn('no_ack', specification.Basic.Get())

    def test_basic_getempty_has_cluster_id(self):
        self.assertIn('cluster_id', specification.Basic.GetEmpty())

    def test_basic_getok_has_delivery_tag(self):
        self.assertIn('delivery_tag', specification.Basic.GetOk())

    def test_basic_getok_has_redelivered(self):
        self.assertIn('redelivered', specification.Basic.GetOk())

    def test_basic_getok_has_exchange(self):
        self.assertIn('exchange', specification.Basic.GetOk())

    def test_basic_getok_has_routing_key(self):
        self.assertIn('routing_key', specification.Basic.GetOk())

    def test_basic_getok_has_message_count(self):
        self.assertIn('message_count', specification.Basic.GetOk())

    def test_basic_nack_has_delivery_tag(self):
        self.assertIn('delivery_tag', specification.Basic.Nack())

    def test_basic_nack_has_multiple(self):
        self.assertIn('multiple', specification.Basic.Nack())

    def test_basic_nack_has_requeue(self):
        self.assertIn('requeue', specification.Basic.Nack())

    def test_basic_publish_has_ticket(self):
        self.assertIn('ticket', specification.Basic.Publish())

    def test_basic_publish_has_exchange(self):
        self.assertIn('exchange', specification.Basic.Publish())

    def test_basic_publish_has_routing_key(self):
        self.assertIn('routing_key', specification.Basic.Publish())

    def test_basic_publish_has_mandatory(self):
        self.assertIn('mandatory', specification.Basic.Publish())

    def test_basic_publish_has_immediate(self):
        self.assertIn('immediate', specification.Basic.Publish())

    def test_basic_qos_has_prefetch_size(self):
        self.assertIn('prefetch_size', specification.Basic.Qos())

    def test_basic_qos_has_prefetch_count(self):
        self.assertIn('prefetch_count', specification.Basic.Qos())

    def test_basic_qos_has_global_(self):
        self.assertIn('global_', specification.Basic.Qos())

    def test_basic_recover_has_requeue(self):
        self.assertIn('requeue', specification.Basic.Recover())

    def test_basic_reject_has_delivery_tag(self):
        self.assertIn('delivery_tag', specification.Basic.Reject())

    def test_basic_reject_has_requeue(self):
        self.assertIn('requeue', specification.Basic.Reject())

    def test_basic_return_has_reply_code(self):
        self.assertIn('reply_code', specification.Basic.Return())

    def test_basic_return_has_reply_text(self):
        self.assertIn('reply_text', specification.Basic.Return())

    def test_basic_return_has_exchange(self):
        self.assertIn('exchange', specification.Basic.Return())

    def test_basic_return_has_routing_key(self):
        self.assertIn('routing_key', specification.Basic.Return())

    def test_channel_close_has_reply_code(self):
        self.assertIn('reply_code', specification.Channel.Close())

    def test_channel_close_has_reply_text(self):
        self.assertIn('reply_text', specification.Channel.Close())

    def test_channel_close_has_class_id(self):
        self.assertIn('class_id', specification.Channel.Close())

    def test_channel_close_has_method_id(self):
        self.assertIn('method_id', specification.Channel.Close())

    def test_channel_flow_has_active(self):
        self.assertIn('active', specification.Channel.Flow())

    def test_channel_flowok_has_active(self):
        self.assertIn('active', specification.Channel.FlowOk())

    def test_channel_open_has_out_of_band(self):
        self.assertIn('out_of_band', specification.Channel.Open())

    def test_channel_openok_has_channel_id(self):
        self.assertIn('channel_id', specification.Channel.OpenOk())

    def test_confirm_select_has_nowait(self):
        self.assertIn('nowait', specification.Confirm.Select())

    def test_connection_blocked_has_reason(self):
        self.assertIn('reason', specification.Connection.Blocked())

    def test_connection_close_has_reply_code(self):
        self.assertIn('reply_code', specification.Connection.Close())

    def test_connection_close_has_reply_text(self):
        self.assertIn('reply_text', specification.Connection.Close())

    def test_connection_close_has_class_id(self):
        self.assertIn('class_id', specification.Connection.Close())

    def test_connection_close_has_method_id(self):
        self.assertIn('method_id', specification.Connection.Close())

    def test_connection_open_has_virtual_host(self):
        self.assertIn('virtual_host', specification.Connection.Open())

    def test_connection_open_has_capabilities(self):
        self.assertIn('capabilities', specification.Connection.Open())

    def test_connection_open_has_insist(self):
        self.assertIn('insist', specification.Connection.Open())

    def test_connection_openok_has_known_hosts(self):
        self.assertIn('known_hosts', specification.Connection.OpenOk())

    def test_connection_secure_has_challenge(self):
        self.assertIn('challenge', specification.Connection.Secure())

    def test_connection_secureok_has_response(self):
        self.assertIn('response', specification.Connection.SecureOk())

    def test_connection_start_has_version_major(self):
        self.assertIn('version_major', specification.Connection.Start())

    def test_connection_start_has_version_minor(self):
        self.assertIn('version_minor', specification.Connection.Start())

    def test_connection_start_has_server_properties(self):
        self.assertIn('server_properties', specification.Connection.Start())

    def test_connection_start_has_mechanisms(self):
        self.assertIn('mechanisms', specification.Connection.Start())

    def test_connection_start_has_locales(self):
        self.assertIn('locales', specification.Connection.Start())

    def test_connection_startok_has_mechanism(self):
        self.assertIn('mechanism', specification.Connection.StartOk())

    def test_connection_startok_has_response(self):
        self.assertIn('response', specification.Connection.StartOk())

    def test_connection_startok_has_locale(self):
        self.assertIn('locale', specification.Connection.StartOk())

    def test_connection_tune_has_channel_max(self):
        self.assertIn('channel_max', specification.Connection.Tune())

    def test_connection_tune_has_frame_max(self):
        self.assertIn('frame_max', specification.Connection.Tune())

    def test_connection_tune_has_heartbeat(self):
        self.assertIn('heartbeat', specification.Connection.Tune())

    def test_connection_tuneok_has_channel_max(self):
        self.assertIn('channel_max', specification.Connection.TuneOk())

    def test_connection_tuneok_has_frame_max(self):
        self.assertIn('frame_max', specification.Connection.TuneOk())

    def test_connection_tuneok_has_heartbeat(self):
        self.assertIn('heartbeat', specification.Connection.TuneOk())

    def test_exchange_bind_has_ticket(self):
        self.assertIn('ticket', specification.Exchange.Bind())

    def test_exchange_bind_has_destination(self):
        self.assertIn('destination', specification.Exchange.Bind())

    def test_exchange_bind_has_source(self):
        self.assertIn('source', specification.Exchange.Bind())

    def test_exchange_bind_has_routing_key(self):
        self.assertIn('routing_key', specification.Exchange.Bind())

    def test_exchange_bind_has_nowait(self):
        self.assertIn('nowait', specification.Exchange.Bind())

    def test_exchange_bind_has_arguments(self):
        self.assertIn('arguments', specification.Exchange.Bind())

    def test_exchange_declare_has_ticket(self):
        self.assertIn('ticket', specification.Exchange.Declare())

    def test_exchange_declare_has_exchange(self):
        self.assertIn('exchange', specification.Exchange.Declare())

    def test_exchange_declare_has_exchange_type(self):
        self.assertIn('exchange_type', specification.Exchange.Declare())

    def test_exchange_declare_has_passive(self):
        self.assertIn('passive', specification.Exchange.Declare())

    def test_exchange_declare_has_durable(self):
        self.assertIn('durable', specification.Exchange.Declare())

    def test_exchange_declare_has_auto_delete(self):
        self.assertIn('auto_delete', specification.Exchange.Declare())

    def test_exchange_declare_has_internal(self):
        self.assertIn('internal', specification.Exchange.Declare())

    def test_exchange_declare_has_nowait(self):
        self.assertIn('nowait', specification.Exchange.Declare())

    def test_exchange_declare_has_arguments(self):
        self.assertIn('arguments', specification.Exchange.Declare())

    def test_exchange_delete_has_ticket(self):
        self.assertIn('ticket', specification.Exchange.Delete())

    def test_exchange_delete_has_exchange(self):
        self.assertIn('exchange', specification.Exchange.Delete())

    def test_exchange_delete_has_if_unused(self):
        self.assertIn('if_unused', specification.Exchange.Delete())

    def test_exchange_delete_has_nowait(self):
        self.assertIn('nowait', specification.Exchange.Delete())

    def test_exchange_unbind_has_ticket(self):
        self.assertIn('ticket', specification.Exchange.Unbind())

    def test_exchange_unbind_has_destination(self):
        self.assertIn('destination', specification.Exchange.Unbind())

    def test_exchange_unbind_has_source(self):
        self.assertIn('source', specification.Exchange.Unbind())

    def test_exchange_unbind_has_routing_key(self):
        self.assertIn('routing_key', specification.Exchange.Unbind())

    def test_exchange_unbind_has_nowait(self):
        self.assertIn('nowait', specification.Exchange.Unbind())

    def test_exchange_unbind_has_arguments(self):
        self.assertIn('arguments', specification.Exchange.Unbind())

    def test_queue_bind_has_ticket(self):
        self.assertIn('ticket', specification.Queue.Bind())

    def test_queue_bind_has_queue(self):
        self.assertIn('queue', specification.Queue.Bind())

    def test_queue_bind_has_exchange(self):
        self.assertIn('exchange', specification.Queue.Bind())

    def test_queue_bind_has_routing_key(self):
        self.assertIn('routing_key', specification.Queue.Bind())

    def test_queue_bind_has_nowait(self):
        self.assertIn('nowait', specification.Queue.Bind())

    def test_queue_bind_has_arguments(self):
        self.assertIn('arguments', specification.Queue.Bind())

    def test_queue_declare_has_ticket(self):
        self.assertIn('ticket', specification.Queue.Declare())

    def test_queue_declare_has_queue(self):
        self.assertIn('queue', specification.Queue.Declare())

    def test_queue_declare_has_passive(self):
        self.assertIn('passive', specification.Queue.Declare())

    def test_queue_declare_has_durable(self):
        self.assertIn('durable', specification.Queue.Declare())

    def test_queue_declare_has_exclusive(self):
        self.assertIn('exclusive', specification.Queue.Declare())

    def test_queue_declare_has_auto_delete(self):
        self.assertIn('auto_delete', specification.Queue.Declare())

    def test_queue_declare_has_nowait(self):
        self.assertIn('nowait', specification.Queue.Declare())

    def test_queue_declare_has_arguments(self):
        self.assertIn('arguments', specification.Queue.Declare())

    def test_queue_declareok_has_queue(self):
        self.assertIn('queue', specification.Queue.DeclareOk())

    def test_queue_declareok_has_message_count(self):
        self.assertIn('message_count', specification.Queue.DeclareOk())

    def test_queue_declareok_has_consumer_count(self):
        self.assertIn('consumer_count', specification.Queue.DeclareOk())

    def test_queue_delete_has_ticket(self):
        self.assertIn('ticket', specification.Queue.Delete())

    def test_queue_delete_has_queue(self):
        self.assertIn('queue', specification.Queue.Delete())

    def test_queue_delete_has_if_unused(self):
        self.assertIn('if_unused', specification.Queue.Delete())

    def test_queue_delete_has_if_empty(self):
        self.assertIn('if_empty', specification.Queue.Delete())

    def test_queue_delete_has_nowait(self):
        self.assertIn('nowait', specification.Queue.Delete())

    def test_queue_deleteok_has_message_count(self):
        self.assertIn('message_count', specification.Queue.DeleteOk())

    def test_queue_purge_has_ticket(self):
        self.assertIn('ticket', specification.Queue.Purge())

    def test_queue_purge_has_queue(self):
        self.assertIn('queue', specification.Queue.Purge())

    def test_queue_purge_has_nowait(self):
        self.assertIn('nowait', specification.Queue.Purge())

    def test_queue_purgeok_has_message_count(self):
        self.assertIn('message_count', specification.Queue.PurgeOk())

    def test_queue_unbind_has_ticket(self):
        self.assertIn('ticket', specification.Queue.Unbind())

    def test_queue_unbind_has_queue(self):
        self.assertIn('queue', specification.Queue.Unbind())

    def test_queue_unbind_has_exchange(self):
        self.assertIn('exchange', specification.Queue.Unbind())

    def test_queue_unbind_has_routing_key(self):
        self.assertIn('routing_key', specification.Queue.Unbind())

    def test_queue_unbind_has_arguments(self):
        self.assertIn('arguments', specification.Queue.Unbind())


class DeprecationWarningTests(unittest.TestCase):
    def test_basic_recoverasync_raises_deprecation_error(self):
        with self.assertWarns(DeprecationWarning):
            specification.Basic.RecoverAsync()


class BasicPropertiesTests(unittest.TestCase):
    def test_basic_properties_has_content_type(self):
        self.assertEqual(specification.Basic.Properties.type('content_type'),
                         'shortstr')

    def test_basic_properties_has_content_encoding(self):
        self.assertEqual(
            specification.Basic.Properties.type('content_encoding'),
            'shortstr')

    def test_basic_properties_has_headers(self):
        self.assertEqual(specification.Basic.Properties.type('headers'),
                         'table')

    def test_basic_properties_has_delivery_mode(self):
        self.assertEqual(specification.Basic.Properties.type('delivery_mode'),
                         'octet')

    def test_basic_properties_has_priority(self):
        self.assertEqual(specification.Basic.Properties.type('priority'),
                         'octet')

    def test_basic_properties_has_correlation_id(self):
        self.assertEqual(specification.Basic.Properties.type('correlation_id'),
                         'shortstr')

    def test_basic_properties_has_reply_to(self):
        self.assertEqual(specification.Basic.Properties.type('reply_to'),
                         'shortstr')

    def test_basic_properties_has_expiration(self):
        self.assertEqual(specification.Basic.Properties.type('expiration'),
                         'shortstr')

    def test_basic_properties_has_message_id(self):
        self.assertEqual(specification.Basic.Properties.type('message_id'),
                         'shortstr')

    def test_basic_properties_has_timestamp(self):
        self.assertEqual(specification.Basic.Properties.type('timestamp'),
                         'timestamp')

    def test_basic_properties_has_message_type(self):
        self.assertEqual(specification.Basic.Properties.type('message_type'),
                         'shortstr')

    def test_basic_properties_has_user_id(self):
        self.assertEqual(specification.Basic.Properties.type('user_id'),
                         'shortstr')

    def test_basic_properties_has_app_id(self):
        self.assertEqual(specification.Basic.Properties.type('app_id'),
                         'shortstr')

    def test_basic_properties_has_cluster_id(self):
        self.assertEqual(specification.Basic.Properties.type('cluster_id'),
                         'shortstr')


class MethodAttributeLengthTests(unittest.TestCase):
    def test_basic_ack_attribute_count(self):
        self.assertEqual(len(specification.Basic.Ack()), 2)

    def test_basic_cancel_attribute_count(self):
        self.assertEqual(len(specification.Basic.Cancel()), 2)

    def test_basic_cancelok_attribute_count(self):
        self.assertEqual(len(specification.Basic.CancelOk()), 1)

    def test_basic_consume_attribute_count(self):
        self.assertEqual(len(specification.Basic.Consume()), 8)

    def test_basic_consumeok_attribute_count(self):
        self.assertEqual(len(specification.Basic.ConsumeOk()), 1)

    def test_basic_deliver_attribute_count(self):
        self.assertEqual(len(specification.Basic.Deliver()), 5)

    def test_basic_get_attribute_count(self):
        self.assertEqual(len(specification.Basic.Get()), 3)

    def test_basic_getempty_attribute_count(self):
        self.assertEqual(len(specification.Basic.GetEmpty()), 1)

    def test_basic_getok_attribute_count(self):
        self.assertEqual(len(specification.Basic.GetOk()), 5)

    def test_basic_nack_attribute_count(self):
        self.assertEqual(len(specification.Basic.Nack()), 3)

    def test_basic_publish_attribute_count(self):
        self.assertEqual(len(specification.Basic.Publish()), 5)

    def test_basic_qos_attribute_count(self):
        self.assertEqual(len(specification.Basic.Qos()), 3)

    def test_basic_qosok_attribute_count(self):
        self.assertEqual(len(specification.Basic.QosOk()), 0)

    def test_basic_recover_attribute_count(self):
        self.assertEqual(len(specification.Basic.Recover()), 1)

    def test_basic_recoverok_attribute_count(self):
        self.assertEqual(len(specification.Basic.RecoverOk()), 0)

    def test_basic_reject_attribute_count(self):
        self.assertEqual(len(specification.Basic.Reject()), 2)

    def test_basic_return_attribute_count(self):
        self.assertEqual(len(specification.Basic.Return()), 4)

    def test_channel_close_attribute_count(self):
        self.assertEqual(len(specification.Channel.Close()), 4)

    def test_channel_closeok_attribute_count(self):
        self.assertEqual(len(specification.Channel.CloseOk()), 0)

    def test_channel_flow_attribute_count(self):
        self.assertEqual(len(specification.Channel.Flow()), 1)

    def test_channel_flowok_attribute_count(self):
        self.assertEqual(len(specification.Channel.FlowOk()), 1)

    def test_channel_open_attribute_count(self):
        self.assertEqual(len(specification.Channel.Open()), 1)

    def test_channel_openok_attribute_count(self):
        self.assertEqual(len(specification.Channel.OpenOk()), 1)

    def test_confirm_select_attribute_count(self):
        self.assertEqual(len(specification.Confirm.Select()), 1)

    def test_confirm_selectok_attribute_count(self):
        self.assertEqual(len(specification.Confirm.SelectOk()), 0)

    def test_connection_blocked_attribute_count(self):
        self.assertEqual(len(specification.Connection.Blocked()), 1)

    def test_connection_close_attribute_count(self):
        self.assertEqual(len(specification.Connection.Close()), 4)

    def test_connection_closeok_attribute_count(self):
        self.assertEqual(len(specification.Connection.CloseOk()), 0)

    def test_connection_open_attribute_count(self):
        self.assertEqual(len(specification.Connection.Open()), 3)

    def test_connection_openok_attribute_count(self):
        self.assertEqual(len(specification.Connection.OpenOk()), 1)

    def test_connection_secure_attribute_count(self):
        self.assertEqual(len(specification.Connection.Secure()), 1)

    def test_connection_secureok_attribute_count(self):
        self.assertEqual(len(specification.Connection.SecureOk()), 1)

    def test_connection_start_attribute_count(self):
        self.assertEqual(len(specification.Connection.Start()), 5)

    def test_connection_startok_attribute_count(self):
        self.assertEqual(len(specification.Connection.StartOk()), 4)

    def test_connection_tune_attribute_count(self):
        self.assertEqual(len(specification.Connection.Tune()), 3)

    def test_connection_tuneok_attribute_count(self):
        self.assertEqual(len(specification.Connection.TuneOk()), 3)

    def test_connection_unblocked_attribute_count(self):
        self.assertEqual(len(specification.Connection.Unblocked()), 0)

    def test_exchange_bind_attribute_count(self):
        self.assertEqual(len(specification.Exchange.Bind()), 6)

    def test_exchange_bindok_attribute_count(self):
        self.assertEqual(len(specification.Exchange.BindOk()), 0)

    def test_exchange_declare_attribute_count(self):
        self.assertEqual(len(specification.Exchange.Declare()), 9)

    def test_exchange_declareok_attribute_count(self):
        self.assertEqual(len(specification.Exchange.DeclareOk()), 0)

    def test_exchange_delete_attribute_count(self):
        self.assertEqual(len(specification.Exchange.Delete()), 4)

    def test_exchange_deleteok_attribute_count(self):
        self.assertEqual(len(specification.Exchange.DeleteOk()), 0)

    def test_exchange_unbind_attribute_count(self):
        self.assertEqual(len(specification.Exchange.Unbind()), 6)

    def test_exchange_unbindok_attribute_count(self):
        self.assertEqual(len(specification.Exchange.UnbindOk()), 0)

    def test_queue_bind_attribute_count(self):
        self.assertEqual(len(specification.Queue.Bind()), 6)

    def test_queue_bindok_attribute_count(self):
        self.assertEqual(len(specification.Queue.BindOk()), 0)

    def test_queue_declare_attribute_count(self):
        self.assertEqual(len(specification.Queue.Declare()), 8)

    def test_queue_declareok_attribute_count(self):
        self.assertEqual(len(specification.Queue.DeclareOk()), 3)

    def test_queue_delete_attribute_count(self):
        self.assertEqual(len(specification.Queue.Delete()), 5)

    def test_queue_deleteok_attribute_count(self):
        self.assertEqual(len(specification.Queue.DeleteOk()), 1)

    def test_queue_purge_attribute_count(self):
        self.assertEqual(len(specification.Queue.Purge()), 3)

    def test_queue_purgeok_attribute_count(self):
        self.assertEqual(len(specification.Queue.PurgeOk()), 1)

    def test_queue_unbind_attribute_count(self):
        self.assertEqual(len(specification.Queue.Unbind()), 5)

    def test_queue_unbindok_attribute_count(self):
        self.assertEqual(len(specification.Queue.UnbindOk()), 0)

    def test_tx_commit_attribute_count(self):
        self.assertEqual(len(specification.Tx.Commit()), 0)

    def test_tx_commitok_attribute_count(self):
        self.assertEqual(len(specification.Tx.CommitOk()), 0)

    def test_tx_rollback_attribute_count(self):
        self.assertEqual(len(specification.Tx.Rollback()), 0)

    def test_tx_rollbackok_attribute_count(self):
        self.assertEqual(len(specification.Tx.RollbackOk()), 0)

    def test_tx_select_attribute_count(self):
        self.assertEqual(len(specification.Tx.Select()), 0)

    def test_tx_selectok_attribute_count(self):
        self.assertEqual(len(specification.Tx.SelectOk()), 0)


class MethodAttributeDefaultTests(unittest.TestCase):
    def test_basic_ack_default_for_delivery_tag(self):
        obj = specification.Basic.Ack()
        self.assertEqual(obj['delivery_tag'], 0)

    def test_basic_ack_default_for_multiple(self):
        obj = specification.Basic.Ack()
        self.assertEqual(obj['multiple'], False)

    def test_basic_cancel_default_for_consumer_tag(self):
        obj = specification.Basic.Cancel()
        self.assertEqual(obj['consumer_tag'], '')

    def test_basic_cancel_default_for_nowait(self):
        obj = specification.Basic.Cancel()
        self.assertEqual(obj['nowait'], False)

    def test_basic_cancelok_default_for_consumer_tag(self):
        obj = specification.Basic.CancelOk()
        self.assertEqual(obj['consumer_tag'], '')

    def test_basic_consume_default_for_ticket(self):
        obj = specification.Basic.Consume()
        self.assertEqual(obj['ticket'], 0)

    def test_basic_consume_default_for_queue(self):
        obj = specification.Basic.Consume()
        self.assertEqual(obj['queue'], '')

    def test_basic_consume_default_for_consumer_tag(self):
        obj = specification.Basic.Consume()
        self.assertEqual(obj['consumer_tag'], '')

    def test_basic_consume_default_for_no_local(self):
        obj = specification.Basic.Consume()
        self.assertEqual(obj['no_local'], False)

    def test_basic_consume_default_for_no_ack(self):
        obj = specification.Basic.Consume()
        self.assertEqual(obj['no_ack'], False)

    def test_basic_consume_default_for_exclusive(self):
        obj = specification.Basic.Consume()
        self.assertEqual(obj['exclusive'], False)

    def test_basic_consume_default_for_nowait(self):
        obj = specification.Basic.Consume()
        self.assertEqual(obj['nowait'], False)

    def test_basic_consume_default_for_arguments(self):
        obj = specification.Basic.Consume()
        self.assertDictEqual(obj['arguments'], {})

    def test_basic_consumeok_default_for_consumer_tag(self):
        obj = specification.Basic.ConsumeOk()
        self.assertEqual(obj['consumer_tag'], '')

    def test_basic_deliver_default_for_consumer_tag(self):
        obj = specification.Basic.Deliver()
        self.assertEqual(obj['consumer_tag'], '')

    def test_basic_deliver_default_for_delivery_tag(self):
        obj = specification.Basic.Deliver()
        self.assertEqual(obj['delivery_tag'], None)

    def test_basic_deliver_default_for_redelivered(self):
        obj = specification.Basic.Deliver()
        self.assertEqual(obj['redelivered'], False)

    def test_basic_deliver_default_for_exchange(self):
        obj = specification.Basic.Deliver()
        self.assertEqual(obj['exchange'], '')

    def test_basic_deliver_default_for_routing_key(self):
        obj = specification.Basic.Deliver()
        self.assertEqual(obj['routing_key'], '')

    def test_basic_get_default_for_ticket(self):
        obj = specification.Basic.Get()
        self.assertEqual(obj['ticket'], 0)

    def test_basic_get_default_for_queue(self):
        obj = specification.Basic.Get()
        self.assertEqual(obj['queue'], '')

    def test_basic_get_default_for_no_ack(self):
        obj = specification.Basic.Get()
        self.assertEqual(obj['no_ack'], False)

    def test_basic_getempty_default_for_cluster_id(self):
        obj = specification.Basic.GetEmpty()
        self.assertEqual(obj['cluster_id'], '')

    def test_basic_getok_default_for_delivery_tag(self):
        obj = specification.Basic.GetOk()
        self.assertEqual(obj['delivery_tag'], None)

    def test_basic_getok_default_for_redelivered(self):
        obj = specification.Basic.GetOk()
        self.assertEqual(obj['redelivered'], False)

    def test_basic_getok_default_for_exchange(self):
        obj = specification.Basic.GetOk()
        self.assertEqual(obj['exchange'], '')

    def test_basic_getok_default_for_routing_key(self):
        obj = specification.Basic.GetOk()
        self.assertEqual(obj['routing_key'], '')

    def test_basic_getok_default_for_message_count(self):
        obj = specification.Basic.GetOk()
        self.assertEqual(obj['message_count'], 0)

    def test_basic_nack_default_for_delivery_tag(self):
        obj = specification.Basic.Nack()
        self.assertEqual(obj['delivery_tag'], 0)

    def test_basic_nack_default_for_multiple(self):
        obj = specification.Basic.Nack()
        self.assertEqual(obj['multiple'], False)

    def test_basic_nack_default_for_requeue(self):
        obj = specification.Basic.Nack()
        self.assertEqual(obj['requeue'], True)

    def test_basic_publish_default_for_ticket(self):
        obj = specification.Basic.Publish()
        self.assertEqual(obj['ticket'], 0)

    def test_basic_publish_default_for_exchange(self):
        obj = specification.Basic.Publish()
        self.assertEqual(obj['exchange'], '')

    def test_basic_publish_default_for_routing_key(self):
        obj = specification.Basic.Publish()
        self.assertEqual(obj['routing_key'], '')

    def test_basic_publish_default_for_mandatory(self):
        obj = specification.Basic.Publish()
        self.assertEqual(obj['mandatory'], False)

    def test_basic_publish_default_for_immediate(self):
        obj = specification.Basic.Publish()
        self.assertEqual(obj['immediate'], False)

    def test_basic_qos_default_for_prefetch_size(self):
        obj = specification.Basic.Qos()
        self.assertEqual(obj['prefetch_size'], 0)

    def test_basic_qos_default_for_prefetch_count(self):
        obj = specification.Basic.Qos()
        self.assertEqual(obj['prefetch_count'], 0)

    def test_basic_qos_default_for_global_(self):
        obj = specification.Basic.Qos()
        self.assertEqual(obj['global_'], False)

    def test_basic_recover_default_for_requeue(self):
        obj = specification.Basic.Recover()
        self.assertEqual(obj['requeue'], False)

    def test_basic_reject_default_for_delivery_tag(self):
        obj = specification.Basic.Reject()
        self.assertEqual(obj['delivery_tag'], None)

    def test_basic_reject_default_for_requeue(self):
        obj = specification.Basic.Reject()
        self.assertEqual(obj['requeue'], True)

    def test_basic_return_default_for_reply_code(self):
        obj = specification.Basic.Return()
        self.assertEqual(obj['reply_code'], 0)

    def test_basic_return_default_for_reply_text(self):
        obj = specification.Basic.Return()
        self.assertEqual(obj['reply_text'], '')

    def test_basic_return_default_for_exchange(self):
        obj = specification.Basic.Return()
        self.assertEqual(obj['exchange'], '')

    def test_basic_return_default_for_routing_key(self):
        obj = specification.Basic.Return()
        self.assertEqual(obj['routing_key'], '')

    def test_channel_close_default_for_reply_code(self):
        obj = specification.Channel.Close()
        self.assertEqual(obj['reply_code'], 0)

    def test_channel_close_default_for_reply_text(self):
        obj = specification.Channel.Close()
        self.assertEqual(obj['reply_text'], '')

    def test_channel_close_default_for_class_id(self):
        obj = specification.Channel.Close()
        self.assertEqual(obj['class_id'], 0)

    def test_channel_close_default_for_method_id(self):
        obj = specification.Channel.Close()
        self.assertEqual(obj['method_id'], 0)

    def test_channel_flow_default_for_active(self):
        obj = specification.Channel.Flow()
        self.assertEqual(obj['active'], None)

    def test_channel_flowok_default_for_active(self):
        obj = specification.Channel.FlowOk()
        self.assertEqual(obj['active'], None)

    def test_channel_open_default_for_out_of_band(self):
        obj = specification.Channel.Open()
        self.assertEqual(obj['out_of_band'], '')

    def test_channel_openok_default_for_channel_id(self):
        obj = specification.Channel.OpenOk()
        self.assertEqual(obj['channel_id'], '')

    def test_confirm_select_default_for_nowait(self):
        obj = specification.Confirm.Select()
        self.assertEqual(obj['nowait'], False)

    def test_connection_blocked_default_for_reason(self):
        obj = specification.Connection.Blocked()
        self.assertEqual(obj['reason'], '')

    def test_connection_close_default_for_reply_code(self):
        obj = specification.Connection.Close()
        self.assertEqual(obj['reply_code'], 0)

    def test_connection_close_default_for_reply_text(self):
        obj = specification.Connection.Close()
        self.assertEqual(obj['reply_text'], '')

    def test_connection_close_default_for_class_id(self):
        obj = specification.Connection.Close()
        self.assertEqual(obj['class_id'], 0)

    def test_connection_close_default_for_method_id(self):
        obj = specification.Connection.Close()
        self.assertEqual(obj['method_id'], 0)

    def test_connection_open_default_for_virtual_host(self):
        obj = specification.Connection.Open()
        self.assertEqual(obj['virtual_host'], '/')

    def test_connection_open_default_for_capabilities(self):
        obj = specification.Connection.Open()
        self.assertEqual(obj['capabilities'], '')

    def test_connection_open_default_for_insist(self):
        obj = specification.Connection.Open()
        self.assertEqual(obj['insist'], False)

    def test_connection_openok_default_for_known_hosts(self):
        obj = specification.Connection.OpenOk()
        self.assertEqual(obj['known_hosts'], '')

    def test_connection_secure_default_for_challenge(self):
        obj = specification.Connection.Secure()
        self.assertEqual(obj['challenge'], '')

    def test_connection_secureok_default_for_response(self):
        obj = specification.Connection.SecureOk()
        self.assertEqual(obj['response'], '')

    def test_connection_start_default_for_version_major(self):
        obj = specification.Connection.Start()
        self.assertEqual(obj['version_major'], 0)

    def test_connection_start_default_for_version_minor(self):
        obj = specification.Connection.Start()
        self.assertEqual(obj['version_minor'], 9)

    def test_connection_start_default_for_mechanisms(self):
        obj = specification.Connection.Start()
        self.assertEqual(obj['mechanisms'], 'PLAIN')

    def test_connection_start_default_for_locales(self):
        obj = specification.Connection.Start()
        self.assertEqual(obj['locales'], 'en_US')

    def test_connection_startok_default_for_mechanism(self):
        obj = specification.Connection.StartOk()
        self.assertEqual(obj['mechanism'], 'PLAIN')

    def test_connection_startok_default_for_response(self):
        obj = specification.Connection.StartOk()
        self.assertEqual(obj['response'], '')

    def test_connection_startok_default_for_locale(self):
        obj = specification.Connection.StartOk()
        self.assertEqual(obj['locale'], 'en_US')

    def test_connection_tune_default_for_channel_max(self):
        obj = specification.Connection.Tune()
        self.assertEqual(obj['channel_max'], 0)

    def test_connection_tune_default_for_frame_max(self):
        obj = specification.Connection.Tune()
        self.assertEqual(obj['frame_max'], 0)

    def test_connection_tune_default_for_heartbeat(self):
        obj = specification.Connection.Tune()
        self.assertEqual(obj['heartbeat'], 0)

    def test_connection_tuneok_default_for_channel_max(self):
        obj = specification.Connection.TuneOk()
        self.assertEqual(obj['channel_max'], 0)

    def test_connection_tuneok_default_for_frame_max(self):
        obj = specification.Connection.TuneOk()
        self.assertEqual(obj['frame_max'], 0)

    def test_connection_tuneok_default_for_heartbeat(self):
        obj = specification.Connection.TuneOk()
        self.assertEqual(obj['heartbeat'], 0)

    def test_exchange_bind_default_for_ticket(self):
        obj = specification.Exchange.Bind()
        self.assertEqual(obj['ticket'], 0)

    def test_exchange_bind_default_for_destination(self):
        obj = specification.Exchange.Bind()
        self.assertEqual(obj['destination'], '')

    def test_exchange_bind_default_for_source(self):
        obj = specification.Exchange.Bind()
        self.assertEqual(obj['source'], '')

    def test_exchange_bind_default_for_routing_key(self):
        obj = specification.Exchange.Bind()
        self.assertEqual(obj['routing_key'], '')

    def test_exchange_bind_default_for_nowait(self):
        obj = specification.Exchange.Bind()
        self.assertEqual(obj['nowait'], False)

    def test_exchange_bind_default_for_arguments(self):
        obj = specification.Exchange.Bind()
        self.assertDictEqual(obj['arguments'], {})

    def test_exchange_declare_default_for_ticket(self):
        obj = specification.Exchange.Declare()
        self.assertEqual(obj['ticket'], 0)

    def test_exchange_declare_default_for_exchange(self):
        obj = specification.Exchange.Declare()
        self.assertEqual(obj['exchange'], '')

    def test_exchange_declare_default_for_exchange_type(self):
        obj = specification.Exchange.Declare()
        self.assertEqual(obj['exchange_type'], 'direct')

    def test_exchange_declare_default_for_passive(self):
        obj = specification.Exchange.Declare()
        self.assertEqual(obj['passive'], False)

    def test_exchange_declare_default_for_durable(self):
        obj = specification.Exchange.Declare()
        self.assertEqual(obj['durable'], False)

    def test_exchange_declare_default_for_auto_delete(self):
        obj = specification.Exchange.Declare()
        self.assertEqual(obj['auto_delete'], False)

    def test_exchange_declare_default_for_internal(self):
        obj = specification.Exchange.Declare()
        self.assertEqual(obj['internal'], False)

    def test_exchange_declare_default_for_nowait(self):
        obj = specification.Exchange.Declare()
        self.assertEqual(obj['nowait'], False)

    def test_exchange_declare_default_for_arguments(self):
        obj = specification.Exchange.Declare()
        self.assertDictEqual(obj['arguments'], {})

    def test_exchange_delete_default_for_ticket(self):
        obj = specification.Exchange.Delete()
        self.assertEqual(obj['ticket'], 0)

    def test_exchange_delete_default_for_exchange(self):
        obj = specification.Exchange.Delete()
        self.assertEqual(obj['exchange'], '')

    def test_exchange_delete_default_for_if_unused(self):
        obj = specification.Exchange.Delete()
        self.assertEqual(obj['if_unused'], False)

    def test_exchange_delete_default_for_nowait(self):
        obj = specification.Exchange.Delete()
        self.assertEqual(obj['nowait'], False)

    def test_exchange_unbind_default_for_ticket(self):
        obj = specification.Exchange.Unbind()
        self.assertEqual(obj['ticket'], 0)

    def test_exchange_unbind_default_for_destination(self):
        obj = specification.Exchange.Unbind()
        self.assertEqual(obj['destination'], '')

    def test_exchange_unbind_default_for_source(self):
        obj = specification.Exchange.Unbind()
        self.assertEqual(obj['source'], '')

    def test_exchange_unbind_default_for_routing_key(self):
        obj = specification.Exchange.Unbind()
        self.assertEqual(obj['routing_key'], '')

    def test_exchange_unbind_default_for_nowait(self):
        obj = specification.Exchange.Unbind()
        self.assertEqual(obj['nowait'], False)

    def test_exchange_unbind_default_for_arguments(self):
        obj = specification.Exchange.Unbind()
        self.assertDictEqual(obj['arguments'], {})

    def test_queue_bind_default_for_ticket(self):
        obj = specification.Queue.Bind()
        self.assertEqual(obj['ticket'], 0)

    def test_queue_bind_default_for_queue(self):
        obj = specification.Queue.Bind()
        self.assertEqual(obj['queue'], '')

    def test_queue_bind_default_for_exchange(self):
        obj = specification.Queue.Bind()
        self.assertEqual(obj['exchange'], '')

    def test_queue_bind_default_for_routing_key(self):
        obj = specification.Queue.Bind()
        self.assertEqual(obj['routing_key'], '')

    def test_queue_bind_default_for_nowait(self):
        obj = specification.Queue.Bind()
        self.assertEqual(obj['nowait'], False)

    def test_queue_bind_default_for_arguments(self):
        obj = specification.Queue.Bind()
        self.assertDictEqual(obj['arguments'], {})

    def test_queue_declare_default_for_ticket(self):
        obj = specification.Queue.Declare()
        self.assertEqual(obj['ticket'], 0)

    def test_queue_declare_default_for_queue(self):
        obj = specification.Queue.Declare()
        self.assertEqual(obj['queue'], '')

    def test_queue_declare_default_for_passive(self):
        obj = specification.Queue.Declare()
        self.assertEqual(obj['passive'], False)

    def test_queue_declare_default_for_durable(self):
        obj = specification.Queue.Declare()
        self.assertEqual(obj['durable'], False)

    def test_queue_declare_default_for_exclusive(self):
        obj = specification.Queue.Declare()
        self.assertEqual(obj['exclusive'], False)

    def test_queue_declare_default_for_auto_delete(self):
        obj = specification.Queue.Declare()
        self.assertEqual(obj['auto_delete'], False)

    def test_queue_declare_default_for_nowait(self):
        obj = specification.Queue.Declare()
        self.assertEqual(obj['nowait'], False)

    def test_queue_declare_default_for_arguments(self):
        obj = specification.Queue.Declare()
        self.assertDictEqual(obj['arguments'], {})

    def test_queue_declareok_default_for_queue(self):
        obj = specification.Queue.DeclareOk()
        self.assertEqual(obj['queue'], '')

    def test_queue_declareok_default_for_message_count(self):
        obj = specification.Queue.DeclareOk()
        self.assertEqual(obj['message_count'], 0)

    def test_queue_declareok_default_for_consumer_count(self):
        obj = specification.Queue.DeclareOk()
        self.assertEqual(obj['consumer_count'], 0)

    def test_queue_delete_default_for_ticket(self):
        obj = specification.Queue.Delete()
        self.assertEqual(obj['ticket'], 0)

    def test_queue_delete_default_for_queue(self):
        obj = specification.Queue.Delete()
        self.assertEqual(obj['queue'], '')

    def test_queue_delete_default_for_if_unused(self):
        obj = specification.Queue.Delete()
        self.assertEqual(obj['if_unused'], False)

    def test_queue_delete_default_for_if_empty(self):
        obj = specification.Queue.Delete()
        self.assertEqual(obj['if_empty'], False)

    def test_queue_delete_default_for_nowait(self):
        obj = specification.Queue.Delete()
        self.assertEqual(obj['nowait'], False)

    def test_queue_deleteok_default_for_message_count(self):
        obj = specification.Queue.DeleteOk()
        self.assertEqual(obj['message_count'], 0)

    def test_queue_purge_default_for_ticket(self):
        obj = specification.Queue.Purge()
        self.assertEqual(obj['ticket'], 0)

    def test_queue_purge_default_for_queue(self):
        obj = specification.Queue.Purge()
        self.assertEqual(obj['queue'], '')

    def test_queue_purge_default_for_nowait(self):
        obj = specification.Queue.Purge()
        self.assertEqual(obj['nowait'], False)

    def test_queue_purgeok_default_for_message_count(self):
        obj = specification.Queue.PurgeOk()
        self.assertEqual(obj['message_count'], 0)

    def test_queue_unbind_default_for_ticket(self):
        obj = specification.Queue.Unbind()
        self.assertEqual(obj['ticket'], 0)

    def test_queue_unbind_default_for_queue(self):
        obj = specification.Queue.Unbind()
        self.assertEqual(obj['queue'], '')

    def test_queue_unbind_default_for_exchange(self):
        obj = specification.Queue.Unbind()
        self.assertEqual(obj['exchange'], '')

    def test_queue_unbind_default_for_routing_key(self):
        obj = specification.Queue.Unbind()
        self.assertEqual(obj['routing_key'], '')

    def test_queue_unbind_default_for_arguments(self):
        obj = specification.Queue.Unbind()
        self.assertDictEqual(obj['arguments'], {})



# -*- encoding: utf-8 -*-
"""
AMQP Data Encoder
=================

Functions for encoding data of various types including field tables and arrays

"""
import calendar
import datetime
import decimal as _decimal
import logging
import struct
import time

from pamqp import PYTHON3

LOGGER = logging.getLogger(__name__)

if PYTHON3:
    long = int
    unicode = None  # Dont throw exceptions as it is not defined


DEPRECATED_RABBITMQ_SUPPORT = False
"""Toggle to support older versions of RabbitMQ."""


def support_deprecated_rabbitmq():
    """Invoke to restrict the data types available in field-tables that are
    sent to RabbitMQ.

    """
    global DEPRECATED_RABBITMQ_SUPPORT
    DEPRECATED_RABBITMQ_SUPPORT = True


def bit(value, byte, position):
    """Encode a bit value

    :param int value: Value to decode
    :param int byte: The byte to apply the value to
    :param int position: The position in the byte to set the bit on
    :rtype: tuple of bytes used and a bool value

    """
    return byte | (value << position)


def boolean(value):
    """Encode a boolean value.

    :param bool value: Value to encode
    :rtype: bytes

    """
    if not isinstance(value, bool):
        raise TypeError('bool type required')
    return struct.pack('>B', int(value))


def byte_array(value):
    """Encode a bytearray

    :param bytearray value: Value to encode
    :rtype: bytes

    """
    if not isinstance(value, bytearray):
        raise TypeError('bytearray type required')
    elif PYTHON3:
        return struct.pack('>I', len(value)) + value
    return bytes(struct.pack('>I', len(value)) + value)


def decimal(value):
    """Encode a decimal.Decimal value.

    :param decimal.Decimal value: Value to encode
    :rtype: bytes

    """
    if not isinstance(value, _decimal.Decimal):
        raise TypeError('decimal.Decimal type required')
    tmp = '%s' % value
    if '.' in tmp:
        decimals = len(tmp.split('.')[-1])
        value = value.normalize()
        raw = int(value * (_decimal.Decimal(10)**decimals))
        return struct.pack('>Bi', decimals, raw)
    return struct.pack('>Bi', 0, int(value))


def double(value):
    """Encode a floating point value as a double

    :param float value: Value to encode
    :rtype: str

    """
    if not isinstance(value, float):
        raise TypeError('float type required')
    return struct.pack('>d', value)


def floating_point(value):
    """Encode a floating point value.

    :param float value: Value to encode
    :rtype: bytes

    """
    if not isinstance(value, float):
        raise TypeError('float type required')
    return struct.pack('>f', value)


def long_int(value):
    """Encode a long integer.

    :param int value: Value to encode
    :rtype: bytes

    """
    if PYTHON3 and not isinstance(value, int):
        raise TypeError('int type required')
    elif not isinstance(value, int) and not isinstance(value, long):
        raise TypeError('long type required')
    elif not (-2147483648 <= value <= 2147483647):
        raise TypeError('Long integer range: -2147483648 to 2147483647')
    return struct.pack('>l', value)


def long_uint(value):
    """Encode a long integer.

    :param int value: Value to encode
    :rtype: bytes

    """
    if PYTHON3 and not isinstance(value, int):
        raise TypeError('int type required')
    elif not isinstance(value, int) and not isinstance(value, long):
        raise TypeError('long type required')
    elif not (0 <= value <= 4294967295):
        raise TypeError('Long unsigned-integer range: 0 to 4294967295')
    return struct.pack('>L', value)


def long_long_int(value):
    """Encode a long-long int.

    :param long or int value: Value to encode
    :rtype: bytes

    """
    if PYTHON3 and not isinstance(value, int):
        raise TypeError('int type required')
    elif not isinstance(value, long) and not isinstance(value, int):
        raise TypeError('int or long type required')
    elif not (-9223372036854775808 <= value <= 9223372036854775807):
        raise TypeError('long-long integer range: '
                        '-9223372036854775808 to 9223372036854775807')
    return struct.pack('>q', value)


def long_string(value):
    """Encode a "long string" which the specific defines as any non-null
    data. We will auto-convert str and unicode to bytes but ignore
    bytes objects and place them opaquely.

    :param value: Value to encode
    :type value: bytes or str or unicode
    :rtype: bytes
    :raises: TypeError

    """
    if PYTHON3 and not isinstance(value, (bytes, str)):
        raise TypeError('bytes or str required')
    elif not PYTHON3 and not isinstance(value, (bytes, str, unicode)):
        raise TypeError('bytes, str or unicode required')
    if not isinstance(value, bytes):
        value = _utf8_encode(value)
    return struct.pack('>I', len(value)) + value


def octet(value):
    """Encode an octet value.

    :param value: Value to encode
    :rtype: bytes
    :raises: TypeError

    """
    if not isinstance(value, int):
        raise TypeError('int type required')
    return struct.pack('B', value)


def short_int(value):
    """Encode a short integer.

    :param int value: Value to encode
    :rtype: bytes
    :raises: TypeError

    """
    if not isinstance(value, (int, long)):
        raise TypeError('int or long type required')
    elif not (-32768 <= value <= 32767):
        raise TypeError('Short integer range: -32678 to 32767')
    return struct.pack('>h', value)


def short_uint(value):
    """Encode an unsigned short integer.

    :param int value: Value to encode
    :rtype: bytes
    :raises: TypeError

    """
    if not isinstance(value, int):
        raise TypeError('int type required')
    elif not (0 <= value <= 65535):
        raise TypeError('Short unsigned integer range: 0 to 65535')
    return struct.pack('>H', value)


def short_string(value):
    """ Encode a string.

    :param value: Value to encode
    :type value: bytes or str or unicode
    :rtype: bytes
    :raises: TypeError

    """
    if PYTHON3 and not isinstance(value, (bytes, str)):
        raise TypeError('bytes or str required')
    elif not PYTHON3 and not isinstance(value, (bytes, str, unicode)):
        raise TypeError('bytes, str or unicode required')
    # Ensure that the value is utf-8 encoded if it's unicode
    value = _utf8_encode(value)
    return struct.pack('B', len(value)) + value


def timestamp(value):
    """Encode a datetime.datetime object or time.struct_time.

    :param value: Value to encode
    :type value: datetime.datetime, time.struct_time, integer
    :rtype: bytes
    :raises: TypeError

    """
    if isinstance(value, datetime.datetime):
        value = value.timetuple()
    elif isinstance(value, int):
        value = time.gmtime(value)

    if isinstance(value, time.struct_time):
        return struct.pack('>Q', calendar.timegm(value))
    raise TypeError('datetime.datetime or time.struct_time type required')


def field_array(value):
    """Encode a field array from a dictionary.

    :param list value: Value to encode
    :rtype: bytes
    :raises: TypeError

    """
    if not isinstance(value, list):
        raise TypeError('list type required')

    data = []
    for item in value:
        data.append(encode_table_value(item))

    output = b''.join(data)
    return struct.pack('>I', len(output)) + output


def field_table(value):
    """Encode a field table from a dictionary.

    :param dict or None value: Value to encode
    :rtype: bytes
    :raises: TypeError

    """
    # If there is no value, return a standard 4 null bytes
    if not value:
        return struct.pack('>I', 0)

    elif not isinstance(value, dict):
        raise TypeError('dict type required, got {}'.format(type(value)))

    # Iterate through all of the keys and encode the data into a table
    data = []
    for key, value in sorted(value.items()):

        # UTF-8 encode the key since it behaves like a short-string
        key = _utf8_encode(key)

        # According to the spec, field names should be 128 char max
        if len(key) > 128:
            LOGGER.warning('Truncating key %s to 128 bytes', key)
            key = key[0:128]

        # Append the field header / delimiter
        data.append(struct.pack('B', len(key)))
        data.append(key)
        try:
            data.append(encode_table_value(value))
        except TypeError as err:
            raise TypeError('{} error: {}/'.format(key, err))

    # Join all of the data together as a string
    output = b''.join(data)
    return struct.pack('>I', len(output)) + output


def table_integer(value):
    """Determines the best type of numeric type to encode value as, preferring
    the smallest data size first.

    :param int value: Value to encode
    :rtype: bytes
    :raises: TypeError

    """
    if DEPRECATED_RABBITMQ_SUPPORT:
        return _deprecated_table_integer(value)

    # Send the appropriately sized data value
    if 0 <= value <= 255:
        return b'b' + octet(value)
    elif -32768 <= value <= 32767:
        return b's' + short_int(value)
    elif 0 <= value <= 65535:
        return b'u' + short_uint(value)
    elif -2147483648 <= value <= 2147483647:
        return b'I' + long_int(value)
    elif 0 <= value <= 4294967295:
        return b'i' + long_uint(value)
    elif -9223372036854775808 <= value <= 9223372036854775807:
        return b'l' + long_long_int(value)
    raise TypeError('Unsupported numeric value: {}'.format(value))


def _deprecated_table_integer(value):
    """Determines the best type of numeric type to encode value as, preferring
    the smallest data size first, supporting versions of RabbitMQ < 3.6

    :param int value: Value to encode
    :rtype: bytes
    :raises: TypeError

    """
    # Send the appropriately sized data value
    if 0 <= value <= 255:
        return b'b' + octet(value)
    elif -32768 <= value <= 32767:
        return b's' + short_int(value)
    elif -2147483648 <= value <= 2147483647:
        return b'I' + long_int(value)
    elif -9223372036854775808 <= value <= 9223372036854775807:
        return b'l' + long_long_int(value)
    raise TypeError('Unsupported numeric value: {}'.format(value))


def encode_table_value(value):
    """Takes a value of any type and tries to encode it with the proper encoder

    :param any value: Value to encode
    :rtype: bytes
    :raises: TypeError

    """
    # Determine the field type and encode it
    if isinstance(value, bool):
        result = b't' + boolean(value)

    elif isinstance(value, int) or isinstance(value, long):
        result = table_integer(value)

    elif isinstance(value, _decimal.Decimal):
        result = b'D' + decimal(value)

    elif isinstance(value, float):
        result = b'f' + floating_point(value)

    elif PYTHON3 and isinstance(value, (str, bytes)):
        result = b'S' + long_string(value)

    elif not PYTHON3 and isinstance(value, (str, bytes, unicode)):
        result = b'S' + long_string(value)

    elif (isinstance(value, datetime.datetime) or
          isinstance(value, time.struct_time)):
        result = b'T' + timestamp(value)

    elif isinstance(value, dict):
        result = b'F' + field_table(value)

    elif isinstance(value, list):
        result = b'A' + field_array(value)

    elif isinstance(value, bytearray):
        result = b'x' + byte_array(value)

    elif value is None:
        result = b'V'

    else:
        raise TypeError('Unknown type: {} ({!r})'.format(type(value), value))

    # Return the encoded value
    return result


def by_type(value, data_type):
    """Takes a value of any type and tries to encode it with the specified
    encoder.

    :param any value: Value to encode
    :param str data_type: type of data to encode
    :rtype: bytes
    :raises: TypeError

    """
    # Determine the field type and encode it
    if data_type == 'field_array':
        return field_array(value)
    elif data_type == 'bytearray':
        return byte_array(value)
    elif data_type == 'double':
        return double(value)
    elif data_type == 'long':
        return long_uint(value)
    elif data_type == 'longlong':
        return long_long_int(value)
    elif data_type == 'longstr':
        return long_string(value)
    elif data_type == 'octet':
        return octet(value)
    elif data_type == 'short':
        return short_uint(value)
    elif data_type == 'shortstr':
        return short_string(value)
    elif data_type == 'table':
        return field_table(value)
    elif data_type == 'timestamp':
        return timestamp(value)
    elif data_type == 'void':
        return None
    raise TypeError('Unknown type: {}'.format(value))


def _utf8_encode(value):
    """Ensure that a string or unicode object is UTF-8 encoded.

    :param value: The value to evaluate
    :type value: str, bytes, or unicode
    :rtype: bytes

    """
    if PYTHON3 and isinstance(value, str):
        return value.encode('utf-8')
    elif not PYTHON3 and isinstance(value, unicode):
        return value.encode('utf-8')
    return value




class Frame(object):
    """Base Class for AMQP Methods which specifies the encoding and decoding
    behavior.

    """
    __slots__ = []
    frame_id = 0
    index = 0
    name = 'Frame'
    synchronous = False
    valid_responses = []

    def __iter__(self):
        """Iterate the attributes and values as key, value pairs.

        :rtype: tuple

        """
        for attribute in self.__slots__:
            yield (attribute, getattr(self, attribute))

    def __contains__(self, item):
        """Return if the item is in the attribute list.

        :rtype: bool

        """
        return item in self.__slots__

    def __getitem__(self, item):
        """Return an attribute as if it were a dict.

        :param str item: The item to look for
        :raises: KeyError
        :rtype: any

        """
        return getattr(self, item)

    def __len__(self):
        """Return the length of the attribute list.

        :rtype: int

        """
        return len(self.__slots__)

    def __repr__(self):
        """Return the representation of the frame object

        :return: str

        """
        return '<%s.%s object at %s>' % (__name__, self.name, hex(id(self)))

    @classmethod
    def type(cls, attr):
        """Return the data type for an attribute.

        :rtype: str

        """
        return getattr(cls, '_' + attr)

    def marshal(self):
        """
        Dynamically encode the frame by taking the list of attributes and
        encode them item by item getting the value form the object attribute
        and the data type from the class attribute.

        :rtype: str

        """
        output = []
        processing_bitset = False
        byte = None
        offset = 0
        for argument in self.__slots__:
            data_type = self.type(argument)

            # Check if we need to turn on bit processing
            if not processing_bitset and data_type == 'bit':
                byte = 0
                offset = 0
                processing_bitset = True

            # Get the data value
            data_value = getattr(self, argument)

            # If we're processing a bitset, apply special rules
            if processing_bitset:

                # Is this value not a bit? turn off bitset processing and
                # append the byte value as an octet
                if data_type != 'bit':
                    processing_bitset = False
                    output.append(encode.octet(byte))

                else:
                    # Apply the bit value to the byte
                    byte = encode.bit(data_value, byte, offset)
                    offset += 1
                    if offset == 8:
                        # We've filled a byte for all bits, add the byte
                        output.append(encode.octet(byte))
                        # Turn off processing, we'll turn on in next iteration
                        # if needed
                        processing_bitset = False

                    # Go to the next iteration
                    continue

            # Not a bit, so just process by type
            output.append(encode.by_type(data_value, data_type))

        # Append the last byte if we're processing a bitset
        if processing_bitset:
            output.append(encode.octet(byte))

        return b''.join(output)

    def unmarshal(self, data):
        """
        Dynamically decode the frame data applying the values to the method
        object by iterating through the attributes in order and decoding them.

        :param str data: The binary encoded method data

        """
        offset = 0
        processing_bitset = False
        for argument in self.__slots__:
            data_type = self.type(argument)

            if offset == 7 and processing_bitset:
                data = data[1:]
                offset = 0

            if processing_bitset and data_type != 'bit':
                offset = 0
                processing_bitset = False
                data = data[1:]

            consumed, value = decode.by_type(data, data_type, offset)

            if data_type == 'bit':
                offset += 1
                processing_bitset = True
                consumed = 0

            setattr(self, argument, value)
            if consumed:
                data = data[consumed:]


class PropertiesBase(object):
    """Provide a base object that marshals and unmarshals the Basic.Properties
    object values.

    """

    __slots__ = []
    flags = {}
    name = 'PropertiesBase'

    def __contains__(self, item):
        return item in self.__slots__

    def __delattr__(self, item):
        setattr(self, item, None)

    def __iter__(self):
        """Iterate the attributes and values as key, value pairs.

        :rtype: tuple

        """
        for attribute in self.__slots__:
            yield (attribute, getattr(self, attribute))

    @classmethod
    def attributes(cls):
        """Return the list of attributes

        :rtype: list

        """
        return [attr for attr in cls.__slots__]

    @classmethod
    def type(cls, attr):
        """Return the data type for an attribute.

        :rtype: str

        """
        return getattr(cls, '_' + attr)

    def encode_property(self, property_name, property_value):
        """Encode a single property value

        :param str property_name: The property name to encode
        :param any property_value: The value to encode

        """
        return encode.by_type(property_value, self.type(property_name))

    def marshal(self):
        """Take the Basic.Properties data structure and marshal it into the
        data structure needed for the ContentHeader.

        :rtype: bytes

        """
        flags = 0
        parts = []
        for property_name in self.__slots__:
            property_value = getattr(self, property_name)
            if property_value is not None and property_value != '':
                flags = flags | self.flags[property_name]
                parts.append(self.encode_property(property_name,
                                                  property_value))
        flag_pieces = []
        while True:
            remainder = flags >> 16
            partial_flags = flags & 0xFFFE
            if remainder != 0:
                partial_flags |= 1
            flag_pieces.append(struct.pack('>H', partial_flags))
            flags = remainder
            if not flags:
                break
        return b''.join(flag_pieces + parts)

    def to_dict(self):
        """Return the properties as a dict

        :rtype: dict

        """
        return dict(self)

    def unmarshal(self, flags, data):
        """
        Dynamically decode the frame data applying the values to the method
        object by iterating through the attributes in order and decoding them.

        :param int flags: Flags that indicate if the data has the property
        :param bytes data: The binary encoded method data

        """
        for property_name in self.__slots__:
            if flags & self.flags[property_name]:
                data_type = getattr(self.__class__, '_' + property_name)
                consumed, value = decode.by_type(data, data_type)
                setattr(self, property_name, value)
                data = data[consumed:]



# -*- encoding: utf-8 -*-
"""
pamqp specific exceptions that do not fit within the AMQP exceptions
contained in pamqp.specification

"""


class PAMQPException(Exception):
    """Base exception for all pamqp specific exceptions"""


class UnmarshalingException(PAMQPException):
    """Raised when a frame is not able to be unmarshaled."""

    def __str__(self):
        return 'Could not unmarshal {} frame: {}'.format(
            self.args[0], self.args[1])



# -*- encoding: utf-8 -*-
"""Manage the marshaling and unmarshaling of AMQP frames

unmarshal will turn a raw AMQP byte stream into the appropriate AMQP objects
from the specification file.

marshal will take an object created from the specification file and turn it
into a raw byte stream.

"""
import logging
import struct

from pamqp import (body, decode, exceptions, header, heartbeat, PYTHON3,
                   specification)

AMQP = b'AMQP'
FRAME_HEADER_SIZE = 7
FRAME_END_CHAR = chr(specification.FRAME_END)
DECODE_FRAME_END_CHAR = FRAME_END_CHAR
if PYTHON3:
    FRAME_END_CHAR = bytes((specification.FRAME_END, ))
    DECODE_FRAME_END_CHAR = specification.FRAME_END
LOGGER = logging.getLogger(__name__)
UNMARSHAL_FAILURE = 0, 0, None


def unmarshal(data_in):
    """Takes in binary data and maps builds the appropriate frame type,
    returning a frame object.

    :param bytes data_in: Raw byte stream data
    :rtype: tuple of  bytes consumed, channel, and a frame object
    :raises: specification.FrameError

    """
    # Look to see if it's a protocol header frame
    try:
        frame_value = _unmarshal_protocol_header_frame(data_in)
        if frame_value:
            return 8, 0, frame_value
    except ValueError as error:
        raise exceptions.UnmarshalingException(header.ProtocolHeader, error)

    # Decode the low level frame and break it into parts
    frame_type, channel_id, frame_size = frame_parts(data_in)

    # Heartbeats do not have frame length indicators
    if frame_type == specification.FRAME_HEARTBEAT and frame_size == 0:
        return 8, channel_id, heartbeat.Heartbeat()

    if not frame_size:
        raise exceptions.UnmarshalingException('Unknown', 'No frame size')

    byte_count = FRAME_HEADER_SIZE + frame_size + 1
    if byte_count > len(data_in):
        raise exceptions.UnmarshalingException('Unknown',
                                               'Not all data received')
    if data_in[byte_count - 1] != DECODE_FRAME_END_CHAR:
        raise exceptions.UnmarshalingException('Unknown', 'Last byte error')

    frame_data = data_in[FRAME_HEADER_SIZE:byte_count - 1]

    # Decode a method frame
    if frame_type == specification.FRAME_METHOD:
        return byte_count, channel_id, _unmarshal_method_frame(frame_data)

    # Decode a header frame
    elif frame_type == specification.FRAME_HEADER:
        return byte_count, channel_id, _unmarshal_header_frame(frame_data)

    # Decode a body frame
    elif frame_type == specification.FRAME_BODY:
        return byte_count, channel_id, _unmarshal_body_frame(frame_data)

    raise exceptions.UnmarshalingException(
        'Unknown', 'Unknown frame type: {}'.format(frame_type))


def marshal(frame_value, channel_id):
    """Marshal a frame to be sent over the wire.

    :param frame_value: The frame object to marshal
    :type frame_value: pamqp.specification.Frame or pamqp.heartbeat.Heartbeat
    :param int channel_id: The channel number to send the frame on
    :rtype: str
    :raises: ValueError

    """
    if isinstance(frame_value, header.ProtocolHeader):
        return frame_value.marshal()
    elif isinstance(frame_value, specification.Frame):
        return _marshal_method_frame(frame_value, channel_id)
    elif isinstance(frame_value, header.ContentHeader):
        return _marshal_content_header_frame(frame_value, channel_id)
    elif isinstance(frame_value, body.ContentBody):
        return _marshal_content_body_frame(frame_value, channel_id)
    elif isinstance(frame_value, heartbeat.Heartbeat):
        return frame_value.marshal()
    raise ValueError('Could not determine frame type: {}'.format(frame_value))


def _unmarshal_protocol_header_frame(data_in):
    """Attempt to unmarshal a protocol header frame

    The ProtocolHeader is abbreviated in size and functionality compared to
    the rest of the frame types, so return UNMARSHAL_ERROR doesn't apply
    as cleanly since we don't have all of the attributes to return even
    regardless of success or failure.

    :param bytes data_in: Raw byte stream data
    :rtype: header.ProtocolHeader
    :raises: ValueError

    """
    # Do the first four bytes match?
    if data_in[0:4] == AMQP:
        frame = header.ProtocolHeader()
        frame.unmarshal(data_in)
        return frame


def _unmarshal_method_frame(frame_data):
    """Attempt to unmarshal a method frame

    :param bytes frame_data: Raw frame data to assign to our method frame
    :return tuple: Amount of data consumed and the frame object

    """
    # Get the Method Index from the class data
    bytes_used, method_index = decode.long_int(frame_data[0:4])

    # Create an instance of the method object we're going to unmarshal
    try:
        method = specification.INDEX_MAPPING[method_index]()
    except KeyError:
        raise exceptions.UnmarshalingException(
            'Unknown', 'Unknown method index: {}'.format(str(method_index)))

    # Unmarshal the data
    try:
        method.unmarshal(frame_data[bytes_used:])
    except struct.error as error:
        raise exceptions.UnmarshalingException(method, error)

    #  Unmarshal the data in the object and return it
    return method


def _unmarshal_header_frame(frame_data):
    """Attempt to unmarshal a header frame

    :param bytes frame_data: Raw frame data to assign to our header frame
    :return tuple: Amount of data consumed and the frame object

    """
    content_header = header.ContentHeader()
    try:
        content_header.unmarshal(frame_data)
    except struct.error as error:
        raise exceptions.UnmarshalingException('ContentHeader', error)
    return content_header


def _unmarshal_body_frame(frame_data):
    """Attempt to unmarshal a body frame

    :param bytes frame_data: Raw frame data to assign to our body frame
    :return tuple: Amount of data consumed and the frame object

    """
    content_body = body.ContentBody()
    content_body.unmarshal(frame_data)
    return content_body


def frame_parts(data_in):
    """Try and decode a low-level AMQP frame and return the parts of the frame.

    :param bytes data_in: Raw byte stream data
    :return tuple: frame type, channel number, and frame data to decode

    """
    # Get the Frame Type, Channel Number and Frame Size
    try:
        return struct.unpack('>BHI', data_in[0:FRAME_HEADER_SIZE])
    except struct.error:
        # We didn't get a full frame
        return UNMARSHAL_FAILURE


def _marshal(frame_type, channel_id, payload):
    """Marshal the low-level AMQ frame.

    :param int frame_type: The frame type to marshal
    :param int channel_id: The channel it will be sent on
    :param bytes|bytes payload: The frame payload
    :rtype: str or bytes

    """
    return b''.join([
        struct.pack('>BHI', frame_type, channel_id, len(payload)), payload,
        FRAME_END_CHAR
    ])


def _marshal_content_body_frame(frame_value, channel_id):
    """Marshal as many content body frames as needed to transmit the content

    :param body.ContentBody frame_value: Frame object to marshal
    :param int channel_id: The channel number for the frame(s)
    :rtype: str

    """
    return _marshal(specification.FRAME_BODY, channel_id,
                    frame_value.marshal())


def _marshal_content_header_frame(frame_value, channel_id):
    """Marshal a content header frame

    :param header.ContentHeader frame_value: Frame object to marshal
    :param int channel_id: The channel number for the frame
    :rtype: str

    """
    return _marshal(specification.FRAME_HEADER, channel_id,
                    frame_value.marshal())


def _marshal_method_frame(frame_value, channel_id):
    """Marshal a method frame

    :param specification.Frame frame_value: Frame object to marshal
    :param int channel_id: The channel number for the frame
    :rtype: str

    """
    return _marshal(
        specification.FRAME_METHOD, channel_id,
        struct.pack('>I', frame_value.index) + frame_value.marshal())



import setuptools

from pamqp import __version__

setuptools.setup(
    name='pamqp',
    version=__version__,
    description='RabbitMQ Focused AMQP low-level library',
    long_description=open('README.rst').read(),
    author='Gavin M. Roy',
    author_email='gavinmroy@gmail.com',
    url='http://pamqp.readthedocs.org',
    package_data={'': ['LICENSE', 'README.rst']},
    packages=['pamqp', 'pamqp.codec'],
    extras_require={'codegen': ['lxml']},
    license='BSD',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Topic :: Communications',
        'Topic :: Internet',
        'Topic :: Software Development :: Libraries'
    ],
    zip_safe=True)
